/*-------------------------------------------------------------------------
pam_coda - PAM authentication and password chaning module - Version 0.1
Copyright (C) BORBELY Zoltan, 1999. All rights reserved

Redistribution and use in source and binary forms of Linux-PAM, with
or without modification, are permitted provided that the following
conditions are met:

1. Redistributions of source code must retain any existing copyright
   notice, and this entire permission notice in its entirety,
   including the disclaimer of warranties.

2. Redistributions in binary form must reproduce all prior and current
   copyright notices, this list of conditions, and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.

3. The name of any author may not be used to endorse or promote
   products derived from this software without their specific prior
   written permission.

ALTERNATIVELY, this product may be distributed under the terms of the
GNU General Public License, in which case the provisions of the GNU
GPL are required INSTEAD OF the above restrictions.  (This clause is
necessary due to a potential conflict between the GNU GPL and the
restrictions contained in a BSD-style copyright.)

THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

History:
--------
09/27/1999	0.1	bozo@szivarvanynet.hu	First release

-------------------------------------------------------------------------*/

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <syslog.h>
#include <pwd.h>

#define	PAM_SM_AUTH
#define	PAM_SM_PASSWORD
#include <security/pam_modules.h>

#include <auth2.h>
#include <auser.h>
#include <avenus.h>

#define	MODULE_NAME	"pam_coda"

struct	config
{
	int	debug;
	char	authsrv[256];
};


static void  _pam_log( int err,const char *format,... )
{
	va_list	args;
	
	va_start(args,format);
	openlog(MODULE_NAME,LOG_CONS|LOG_PID,LOG_AUTHPRIV);
	vsyslog(err,format,args);
	va_end(args);
	closelog();
}


static void  get_config( struct config *cfg,int argc,const char *argv[] )
{
	int	i;

	cfg->debug	= 0;
	cfg->authsrv[0]	= '\0';
	for(i=0;i<argc;i++)
	{
		if ( strcmp(argv[i],"debug") == 0 )
		{
			cfg->debug	= 1;
			continue;
		}
		if ( strncmp(argv[i],"authsrv=",8) == 0 )
		{
			if ( strlen(argv[i]+8) <= sizeof(cfg->authsrv) )
			{
				strcpy(cfg->authsrv,argv[i]+8);
			}
			continue;
		}
	}
	if ( cfg->debug )
	{
		_pam_log(LOG_DEBUG,"cfg->debug   = %d\n",cfg->debug);
		_pam_log(LOG_DEBUG,"cfg->authsrv = %s\n",cfg->authsrv);
	}
}


static int  _authenticate(	struct config	*cfg,
				char		*user,
				char		*pass )
{
	EncryptedSecretToken	sToken;
	ClearToken		cToken;
	RPC2_Handle		RPCid;
	RPC2_Integer		c_uid;
	uid_t			s_uid;
	int			retval;

	U_InitRPC();

	retval	= U_BindToServer((cfg->authsrv[0]=='\0') ? NULL : cfg->authsrv,
				 AUTH_METHOD_CODAUSERNAME,
				 user,
				 strlen(user)+1,
				 pass,
				 strlen(pass)+1,
				 &RPCid);
	if ( AuthNameToId(RPCid,user,&c_uid) != 0 )
	{
		_pam_log(LOG_NOTICE,"unknown user (user: %s)",user);
		return( PAM_USER_UNKNOWN );
	}
	if ( retval!=0 )
	{
		_pam_log(LOG_NOTICE,"authentication failure (user: %s)",user);
		return( PAM_AUTH_ERR );
	}

	if ( (retval=AuthGetTokens(RPCid,sToken,&cToken)) != 0 )
	{
		AuthQuit(RPCid);
		_pam_log(LOG_ERR,"AuthGetTokens failed");
		return( PAM_AUTH_ERR );
	}
	RPC2_Unbind(RPCid);

	s_uid	= geteuid();
	seteuid(c_uid);
	retval	= U_SetLocalTokens(0,&cToken,sToken);
	memset(&cToken,0,sizeof(cToken));
	memset(sToken,0,sizeof(sToken));
	seteuid(s_uid);

	if ( retval != 0 )
	{
		_pam_log(LOG_ERR,"U_SetLocalTokens failed");
		return( PAM_AUTH_ERR );
	}

	return( PAM_SUCCESS );
}


static int  _changepass(struct config	*cfg,
			char		*user,
			char		*pass,
			char		*cuser,
			char		*cpass )
{
	int	retval;

	U_InitRPC();

	retval	= U_ChangePassword((cfg->authsrv[0]=='\0') ? NULL : cfg->authsrv,
				   user,
				   pass,
				   AUTH_METHOD_CODAUSERNAME,
				   cuser,
				   strlen(cuser)+1,
				   cpass,
				   strlen(cpass)+1);

	return( (retval == 0) ? PAM_SUCCESS : PAM_AUTH_ERR );
}


PAM_EXTERN int	pam_sm_authenticate(	pam_handle_t	*pamh,
					int		flags,
					int		argc,
					const char	*argv[] )
{
	int		retval;
	char		*user, *pass;
	struct config	cfg;

	get_config(&cfg,argc,argv);

	if ( (retval=pam_get_user(pamh,(const char **)&user,NULL)) != PAM_SUCCESS )
	{
		_pam_log(LOG_ERR,"couldn't determine name of local user");
		return( PAM_AUTH_ERR );
	}
	if ( (retval=pam_get_item(pamh,PAM_AUTHTOK,(const void **)&pass)) != PAM_SUCCESS )
	{
		_pam_log(LOG_ERR,"couldn't get authentication token");
		return( PAM_AUTH_ERR );
	}
	if ( (user==NULL) || (pass==NULL) )
	{
		_pam_log(LOG_ERR,"user name or auth token missing");
		return( PAM_AUTH_ERR );
	}

	return( _authenticate(&cfg,user,pass) );
}


PAM_EXTERN int	pam_sm_setcred(	pam_handle_t	*pamh,
				int		flags,
				int		argc,
				const char	*argv[] )
{
	return( PAM_SUCCESS );
}


PAM_EXTERN int	pam_sm_chauthtok(	pam_handle_t	*pamh,
					int		flags,
					int		argc,
					const char	*argv[] )
{
	int		retval;
	char		*user, *pass, cuser[64], *cpass;
	struct passwd	*pwd;
	struct config	cfg;

	get_config(&cfg,argc,argv);

	if ( !(flags & PAM_UPDATE_AUTHTOK) )
	{
		return( PAM_SUCCESS );
	}

	if ( (retval=pam_get_user(pamh,(const char **)&user,NULL)) != PAM_SUCCESS )
	{
		_pam_log(LOG_ERR,"couldn't determine name of local user");
		return( PAM_AUTH_ERR );
	}

#if	0
	if ( (pwd=getpwuid(getuid())) == NULL )
	{
		_pam_log(LOG_ERR,"couldn't determine name of the invoking user (uid: %lu)",getuid());
		return( PAM_AUTH_ERR );
	}
	if ( strlen(pwd->pw_name) >= sizeof(cuser) )
	{
		_pam_log(LOG_ERR,"invoking user's name is too long (user: %s)",pwd->pw_name);
		return( PAM_AUTH_ERR );
	}
	strcpy(cuser,pwd->pw_name);
#endif

	if ( (retval=pam_get_item(pamh,PAM_AUTHTOK,(const void **)&pass)) != PAM_SUCCESS )
	{
		_pam_log(LOG_ERR,"couldn't get authentication token");
		return( PAM_AUTH_ERR );
	}
	if ( pass == NULL )
	{
		_pam_log(LOG_ERR,"couldn't get authentication token");
		return( PAM_AUTH_ERR );
	}

	if ( (retval=pam_get_item(pamh,PAM_OLDAUTHTOK,(const void **)&cpass)) != PAM_SUCCESS )
	{
		_pam_log(LOG_ERR,"couldn't get old authentication token");
		return( PAM_AUTH_ERR );
	}
	if ( cpass == NULL )
	{
		_pam_log(LOG_ERR,"couldn't get old authentication token");
		return( PAM_AUTH_ERR );
	}

	return( _changepass(&cfg,user,pass,user,cpass) );
//	return( _changepass(&cfg,user,pass,cuser,cpass) );
}


/* static module data */

#ifdef PAM_STATIC
struct pam_module _pam_coda_modstruct = {
	MODULE_NAME,
	pam_sm_authenticate,
	pam_sm_setcred,
	NULL,
	NULL,
	NULL,
	pam_sm_chauthtok
};
#endif
